// ---
// Copyright (c) 2012 Francesco Cottone, http://www.kesiev.com/
// ---
 /*
 * HotGun network client - v0.1
 *
 * USAGE:
 *  HotGun.search("floating duck",function(search){console.log(search.results)});
 *  HotGun.search({query:"floating duck"},function(search){console.log(search.results)});
 *  HotGun.search({query:"floating duck",duration:50000},function(search){console.log(search.results)});
 *
 * REFERENCE:
 *  For HotGun.search({...}) have a look to HotGun._default.
 */
var HotGun={
	config:{
		scriptname:"hotgun.js", // This script name.
		jsonptimeout:5000, // JSONP call timeout (msec)
		cachesize:100, // Caches the first n peers data
		log:false, // Enable logs on console
		shufflepeers:true // Shuffle peers before new search
	},
	_default:{
		query:"", // Default search query
		resultsize:20, // Default number of answers
		duration:10000, // Default duration of a search (msec)
		maxdepth:-1, // Maximum search depth (-1 for no limit)
		onpeer:null, // Callback on new peer contacted for results
		onfilter:null, // Callback for filtering results (built-in regexp matcher function by default)
		onresult:null, // Callback on new result
		overalltime:0 // msec spent searching
	},
	_searchstatus:{
		IN_PROGRESS:0,
		TIMEOUT:1,
		HITSCOUNT_REACHED:2,
		NO_MORE_PEERS:3
	},
	_peerstatus:{
		NOT_FETCHED:0,
		FETCHED:1,
		ERROR:2
	},
	onpeer:null, // Override for logging when receiving data from a new peer (globally)
	_:{resultscount:0,queue:[],peermeta:["label","description"],tag:null,currentcallback:null,currentpeer:null,currenthits:0,currentsearch:null,peerptr:null,timeout:null},
	onlog:function(data){if (this.config.log&&window.console) console.log(data)},
	_shuffle:function(arr) {
		var tmp,pos;
		for (var i=0;i<arr.length;i++) {
			pos=Math.floor(Math.random()*arr.length);
			tmp=arr[pos];
			arr[pos]=arr[i];
			arr[i]=tmp;
		}
		return arr;
	},
	_contexturl:function(url,parent) {
		if (url.indexOf("://")==-1) {
			baseurl=!parent||(parent.indexOf("://")==-1)?document.location.href:parent;
			url=baseurl.substr(0,baseurl.lastIndexOf("/")+1)+url;
		}
		return url;
	},
	_indexmake:function(subj,list){
		for (var i=0;i<list.length;i++) {
			if (!subj[list[i]+"byid"]) subj[list[i]+"byid"]=[];
			if (!subj[list[i]+"byname"]) subj[list[i]+"byname"]={};
		}
		return subj;
	},
	_indexset:function(subj,group,key,value) {
		if (!subj[group+"byname"][key]) {
			subj[group+"byid"].push(value);
			subj[group+"byname"][key]=value;
			return true;
		} else return false;
	},
	_indexmerge:function(subj,group,key,value,refpeer) {
		var data=subj[group+"byname"][key],result;
		if (!data) {
			data={keys:{}};
			this._indexmake(data,["refpeer"]);
		}
		this._indexset(data,"refpeer",refpeer.url,refpeer);
		for (var a in value) {
			if (!data.keys[a]) {
				data.keys[a]={items:[]};
				this._indexmake(data.keys[a],["refpeer"]);
			}
			this._indexset(data.keys[a],"refpeer",refpeer.url,refpeer);
			result=null;
			if ((typeof value[a]=="string")||(typeof value[a]=="number")||(value[a]===null)) for (var i=0;i<data.keys[a].items.length;i++) if (data.keys[a].items[i].value==value[a]) { result=data.keys[a].items[i]; break;}
			if (!result) {
				result={value:value[a]};
				this._indexmake(result,["refpeer"]);
				data.keys[a].items.push(result);
			}
			this._indexset(result,"refpeer",refpeer.url,refpeer);
		}
		if (!subj[group+"byname"][key]) {
			subj[group+"byid"].push(data);
			subj[group+"byname"][key]=data;
			return true;
		} else return false;
	},
	_fetchnext:function() {
		var duration=((new Date).getTime()-this._.currentsearch.searchstarted);
		if (this._.currenthits!==null) this.onlog("( ) "+this._.currenthits+" hit(s) found on ["+this._.currentpeer.url+"]");
		if (duration>this._.currentsearch.duration) {
			this.onlog("(>) Search duration "+duration+"/"+this._.currentsearch.duration+" reached. Ending search.");
			this._fetchend(this._searchstatus.TIMEOUT);
		} else if (this._.currentsearch.resultbyid.length>=this._.currentsearch.resultsize) {
			this.onlog("(>) "+this._.currentsearch.resultbyid.length+"/"+this._.currentsearch.resultsize+" hits reached. Ending search.");
			this._fetchend(this._searchstatus.HITSCOUNT_REACHED);
		} else {
			while (
				this._.peerbyid[this._.peerptr]&&(
					(this._.peerbyid[this._.peerptr].status==this._peerstatus.ERROR)||
					(this._.currentsearch.peerbyname[this._.peerbyid[this._.peerptr].url])||
					((this._.currentsearch.maxdepth>-1)&&(this._.peerbyid[this._.peerptr].depth>this._.currentsearch.maxdepth))
				)
			) this._.peerptr++;
			if (this._.peerbyid[this._.peerptr]) {
				this._.currentpeer=this._.peerbyid[this._.peerptr];
				this._.currenthits=0;
				this.onlog("( ) Updating from ["+this._.currentpeer.url+"]...");
				if (this._.cachebyname[this._.currentpeer.url]) {
					this.onlog("( ) Getting data from cache.");
					var self=this;
					setTimeout(function(){self.feed(self._.cachebyname[self._.currentpeer.url])},1);
				} else {
					this._.tag=document.createElement("script");
					this._.tag.src=this._.currentpeer.url+(this._.currentpeer.url.indexOf("?")==-1?"?":"&")+Math.floor(Math.random()*100000);
					this._.tag.onerror=this._fetcherror;
					this._.timeout=setTimeout(this._fetchtimeout,this.config.jsonptimeout);
					this._.head.appendChild(this._.tag);
				}
			} else {
				this.onlog("(>) No more peers. Ending search.");
				this._fetchend(this._searchstatus.NO_MORE_PEERS);
			}
		}
	},
	_fetchfinalize:function() {
		if (this._.timeout) {
			clearTimeout(this._.timeout);
			this._.timeout=null;
		}
		if (this._.tag) {
			this._.head.removeChild(this._.tag);
			this._.tag=null;
		}
	},
	_fetchtimeout:function() {
		HotGun.onlog("(x) Peer ["+HotGun._.peerbyid[HotGun._.peerptr].url+"] timed out.");
		window.stop();
		HotGun._.timeout=setTimeout(HotGun._fetcherror,10);
	},
	_fetcherror:function() {
		HotGun.onlog("(x) Peer ["+HotGun._.peerbyid[HotGun._.peerptr].url+"] not found.");
		HotGun._fetchfinalize();
		HotGun._.currentpeer.status=HotGun._peerstatus.ERROR;
		if (HotGun.onpeer) HotGun.onpeer(HotGun._.currentpeer,HotGun._.currentsearch,null);
		if (HotGun._indexset(HotGun._.currentsearch,"peer",HotGun._.currentpeer.url,HotGun._.currentpeer)) if (HotGun._.currentsearch.onpeer) HotGun._.currentsearch.onpeer(HotGun._.currentpeer,HotGun._.currentsearch,null);
		HotGun._fetchnext();
	},
	_fetchend:function(status) {
		var sea=this._.currentsearch, cb=this._.currentcallback;
		sea.status=status;
		sea.time=((new Date).getTime()-sea.searchstarted);
		sea.overalltime+=sea.time;
		sea.hasMore=sea.status!=this._searchstatus.NO_MORE_PEERS;
		delete sea.searchstarted;
		this.onlog("(>) Search ["+sea.query+"] ended. "+sea.resultbyid.length+"/"+sea.resultsize+" result(s) in "+sea.time+"msec.");
		if (sea.onsort) sea.onsort(sea);
		if (sea.onpostfilter) sea.onpostfilter(sea);
		sea.results=[];
		for (var i=0;i<sea.resultbyid.length;i++) {
			var itm={};
			for (var a in sea.resultbyid[i].keys) itm[a]=sea.resultbyid[i].keys[a].items[0].value;
			sea.results.push(itm);
		}
		this._.peerptr=null;
		this._.currentsearch=null;
		this._.currentcallback=null;
		if (cb) setTimeout(function(){cb(sea)},1);
		if (this._.queue.length) {
			var next=this._.queue.shift();
			this.onlog("(<) Fetching search ["+next[0].query+"] ("+this._.queue.length+" left)");
			this.search(next[0],next[1]);
		}
	},
	_initialize:function() {
		var scripts=document.getElementsByTagName("SCRIPT");
		this._.head=document.getElementsByTagName('head')[0];
		this._.baseurl=document.location.href;
		for (var i=0;i<scripts.length;i++) {
			if (scripts[i].src&&(scripts[i].src.substr(scripts[i].src.length-this.config.scriptname.length-1,this.config.scriptname.length+1)==("/"+this.config.scriptname))) {
				this._.baseurl=scripts[i].src;
				break;
			}
		}
		this._default.onfilter=this.filters.all;
		this._default.onsort=this.sorts.basic;
		this._default.onpostfilter=this.postFilters.basic;
		this._indexmake(this._,["peer","cache","blacklist"]);
	},
	/*
	 *  Entry point for peers.
	 */
	feed:function(data) {
		var baseurl,peeritems={},peerpeers={},url;
		this._fetchfinalize();
		if (this._.cachebyid.length<this.config.cachesize) this._indexset(this._,"cache",this._.currentpeer.url,data);
		for (var i=0;i<this._.peermeta.length;i++) if (data[this._.peermeta[i]]) this._.currentpeer[this._.peermeta[i]]=data[this._.peermeta[i]];
		if (this._.currentpeer.status==this._peerstatus.NOT_FETCHED) {
			this._.currentpeer.status=this._peerstatus.FETCHED;
			this._.resultscount+=data.items.length;
			peerpeers[this._.currentpeer.url]=1;
			if (data.peers) for (var i=0;i<data.peers.length;i++) {
				data.peers[i]=this._contexturl(data.peers[i],this._.currentpeer.url);
				if (!peerpeers[data.peers[i]]) this.addPeer(data.peers[i],this._.currentpeer);
				peerpeers[data.peers[i]]=1;
			}
			if (data.blacklist) for (var i=0;i<data.blacklist.length;i++) {
				url=this._contexturl(data.blacklist[i],this._.currentpeer.url);
				if (!this._.blacklistbyname[url]) {
					this.onlog("( ) Peer ["+url+"] suggested for blacklisting.");
					this._indexset(this._,"blacklist",url,this._indexmake({},["refpeer"]));
				}
				this._indexset(this._.blacklistbyname[url],"refpeer",this._.currentpeer.url,this._.currentpeer);
			}
			if (this.onpeer) this.onpeer(this._.currentpeer,this._.currentsearch,data);	
		}
		if (this._indexset(this._.currentsearch,"peer",this._.currentpeer.url,this._.currentpeer)) if (this._.currentsearch.onpeer) this._.currentsearch.onpeer(this._.currentpeer,this._.currentsearch,data);
		if (data.items) for (var i=0;i<data.items.length;i++) {
			var item={};
			for (var a in data.items[i]) item[a]=data.items[i][a];
			if (item.url) {
				item.url=this._contexturl(item.url,this._.currentpeer.url);
				item.uid=item.url;
			}
			if (!peeritems[item.uid]&&(this._.currentsearch.onfilter(this._.currentsearch,item,this._.currentpeer,data))) {
				item.peer=this._.currentpeer;
				if (this._indexmerge(this._.currentsearch,"result",item.uid,item,this._.currentpeer)) {
					if (this._.currentsearch.onresult) this._.currentsearch.onresult(item, this._.currentsearch);
					this._.currenthits++;					
				}
			}
			peeritems[item.uid]=1;
		}
		this._fetchnext();
	},
	_regexps:{
		// Reference: http://www.fileinfo.com/filetypes/common
		commonaudio:/(\.aif|\.iff|\.m3u|\.m4a|\.mid|\.mp3|\.mpa|\.ra|\.wav|\.wma)$/i,
		commonvideo:/(\.3g2|\.3gp|\.asf|\.asx|\.avi|\.flv|\.mov|\.mp4|\.mpg|\.rm|\.vob|\.wmv)$/i,
		commonimage:/(\.bmp|\.dds|\.dng|\.gif|\.jpg|\.jpeg|\.png|\.psd|\.tga|\.tif|\.yuv|\.ai|\.eps|\.ps|\.svf)$/i,
		commondocument:/(\.indd|\.pct|\.pdf|\.doc|\.docx|\.log|\.msg|\.odt|\.pages|\.rtf|\.tex|\.txt|\.wpd|\.wps|\.xlr|\.xls|\.xlsx)$/i,
		commonfont:/(\.fnt|\.fon|\.otf|\.ttf)$/i,
		commoncompressed:/(\.7z|\.cbr|\.deb|\.gz|\.pkg|\.rar|\.rpm|\.sit|\.sitx|\.tar.gz|\.zip)$/i,
		audiosites:/(\bsoundcloud\b)/i,
		videosites:/(\byoutube\b|\bvimeo\b)/i
	},
	/*
	 * Public methods and properties
	 */
	filters:{
		url:function(search,item,peer,peerdata){ return item.url && HotGun.filters.all(search,item,peer,peerdata) },
		audio:function(search,item,peer,peerdata){ return item.url && ( (item.type=="audio") || item.url.match(HotGun._regexps.commonaudio) || item.url.match(HotGun._regexps.audiosites) ) && HotGun.filters.all(search,item,peer,peerdata)},
		video:function(search,item,peer,peerdata){ return item.url && ( (item.type=="video") || item.url.match(HotGun._regexps.commonvideo) || item.url.match(HotGun._regexps.videosites) ) && HotGun.filters.all(search,item,peer,peerdata)},
		image:function(search,item,peer,peerdata){ return item.url && ( (item.type=="image") || item.url.match(HotGun._regexps.commonimage) ) && HotGun.filters.all(search,item,peer,peerdata)},
		document:function(search,item,peer,peerdata){ return item.url && ( (item.type=="document") || item.url.match(HotGun._regexps.commondocument) ) && HotGun.filters.all(search,item,peer,peerdata)},
		font:function(search,item,peer,peerdata){ return item.url && ( (item.type=="font") || item.url.match(HotGun._regexps.commonfont) ) && HotGun.filters.all(search,item,peer,peerdata)},
		compressed:function(search,item,peer,peerdata){ return item.url && ( (item.type=="compressed") || item.url.match(HotGun._regexps.commoncompressed) ) && HotGun.filters.all(search,item,peer,peerdata)},
		all:function(currentsearch,currentitem,currentpeer,peerdata) {
			if (!currentsearch.regexpquery) currentsearch.regexpquery=new RegExp(currentsearch.query,"i");
			if (currentsearch.regexpquery) for (var a in currentitem) if ((typeof currentitem[a]=="string")&&currentitem[a].match(currentsearch.regexpquery)) return true;
			return false;
		}
	},
	sortsConfig:{
		basic:{
			onsort:function(a,b) {  // Score sorting callback
				return b.score>a.score ? 1 : -1
			},
			ongetscore:function(list) { // Score calculating
				var score=0;
				for (var peer in list.refpeerbyname)
					if (HotGun._.blacklistbyname[peer]) score+=HotGun._.blacklistbyname[peer].refpeerbyid.length*-2;
					else score+=list.refpeerbyname[peer].refpeerbyid.length*3;
				return score;
			}
		}
	},
	sorts:{
		basic:function(result){
			for (var i=0;i<result.resultbyid.length;i++) {
				for (var key in result.resultbyid[i].keys) {
					for (var s=0;s<result.resultbyid[i].keys[key].items.length;s++)
						result.resultbyid[i].keys[key].items[s].score=HotGun.sortsConfig.basic.ongetscore(result.resultbyid[i].keys[key].items[s]);
					result.resultbyid[i].keys[key].items.sort(HotGun.sortsConfig.basic.onsort);
				}
				result.resultbyid[i].score=HotGun.sortsConfig.basic.ongetscore(result.resultbyid[i]);
			}
			result.resultbyid.sort(HotGun.sortsConfig.basic.onsort);
		}
	},
	postFilters:{
		basic:function(result){
			var i=0,uid;
			while (i<result.resultbyid.length) {
				var keys=0;
				uid=result.resultbyid[i].keys.uid.items[0].value;
				for (var key in result.resultbyid[i].keys) {
					for (var s=0;s<result.resultbyid[i].keys[key].items.length;s++)
						if (result.resultbyid[i].keys[key].items[s].score<0) break;
					if (!s) delete result.resultbyid[i].keys[key]; else {
						result.resultbyid[i].keys[key].items.splice(s,result.resultbyid[i].keys[key].items.length-s);
						keys++;
					}
				}
				if (!keys||!result.resultbyid[i].keys.uid) {
					delete result.resultbyname[uid];
					result.resultbyid.splice(i,1);
				} else i++;
			}
		}
	},
	addPeer:function(url,refpeer) {
		var depth=(refpeer?refpeer.depth+1:0)
		url=this._contexturl(url,(refpeer?refpeer.url:null));
		var newpeer={url:url,status:this._peerstatus.NOT_FETCHED,depth:depth};
		this._indexmake(newpeer,["refpeer"]);
		if (refpeer) this._indexset(newpeer,"refpeer",refpeer.url,refpeer);
		if (this._indexset(this._,"peer",url,newpeer)) {
			this.onlog("(+) Adding peer ["+url+"] ("+depth+" deep)...");
			return true;
		} else return false;
	},
	search:function(env,callback) {
		if (this._.peerbyid.length==0) this.addPeer("peer.js");
		if (typeof env == "string") env={query:env};
		for (var a in this._default) if (!env[a]) env[a]=this._default[a];
		env.searchstarted=(new Date).getTime();
		env.status=this._searchstatus.IN_PROGRESS;
		this._indexmake(env,["peer","result"]);
		if (this._.currentsearch===null) {
			this._.currenthits=null;
			this._.currentsearch=env;
			this._.currentcallback=callback;
			this._.peerptr=0;
			if (this.config.shufflepeers) {
				this.onlog("( ) Shuffling peers...");
				this._shuffle(this._.peerbyid);
			}
			this.onlog("(<) Searching ["+env.query+"]...");
			this._fetchnext();
		} else {
			this.onlog("( ) Search for ["+env.query+"] queued.");
			this._.queue.push([env,callback]);
		}
	}
}
HotGun._initialize();
